"""
``gherlint`` is a linter for Cucumber Gherkin feature files.
"""
from pathlib import Path

import click

from gherlint.fixer import LanguageFixer
from gherlint.linter import GherkinLinter
from gherlint.statistics import compute_metrics


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
)
@click.pass_context
def cli(ctx: click.Context) -> None:
    """gherlint is a linter for Cucumber Gherkin feature files."""
    if ctx.invoked_subcommand is None:
        print(cli.get_help(ctx))


@cli.command()
@click.argument("path")
def lint(path: str) -> None:
    """Perform linting of feature files"""
    GherkinLinter(Path(path)).run()


@cli.command()
@click.argument("path")
def stats(path: str) -> None:
    """Compute metrics over your feature files"""
    compute_metrics(Path(path))


@cli.command()
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Don't write to disk, only output which files would be modified",
)
@click.argument("path")
def fix_language_tags(path: str, dry_run: bool) -> None:
    """Add or fix language tags in feature files

    If gherlint detects that a language other than English is used, it will
    try to infer the correct language and add the corresponding tag.
    If a language tag is present but does not fit to the file contents, the existing
    tag will be replaced.
    """
    LanguageFixer(Path(path)).run(modify=not dry_run)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""Checker focussing on consistency issues."""

from typing import Union

from gherlint.checkers.base_checker import BaseChecker
from gherlint.objectmodel import nodes
from gherlint.registry import CheckerRegistry
from gherlint.reporting import Message


class ConsistencyChecker(BaseChecker):
    MESSAGES = [
        Message(
            "E301",
            "examples-outside-scenario-outline",
            "Examples used outside a Scenario Outline",
        ),
        Message(
            "W301",
            "duplicated-tag",
            "Tag '{tag}' already present on the parent element",
        ),
    ]

    def visit_scenario(
        self, node: Union[nodes.Scenario, nodes.ScenarioOutline]
    ) -> None:
        for example in node.examples:
            self.reporter.add_message("examples-outside-scenario-outline", node=example)
        self._check_duplicated_tag(node)

    def visit_scenariooutline(self, node: nodes.ScenarioOutline) -> None:
        self._check_duplicated_tag(node)

    def visit_examples(self, node: nodes.Examples) -> None:
        self._check_duplicated_tag(node)

    def _check_duplicated_tag(
        self, node: Union[nodes.Scenario, nodes.ScenarioOutline, nodes.Examples]
    ) -> None:
        for tag in node.tags:
            if any(
                tag in parent.tags for parent in node.parents if hasattr(parent, "tags")  # type: ignore
            ):
                self.reporter.add_message("duplicated-tag", node, tag=tag.name)


def register_checker(registry: CheckerRegistry) -> None:
    registry.register(ConsistencyChecker)



from pathlib import Path
from typing import List

import parse

from gherlint.checkers.base_checker import BaseChecker
from gherlint.parser import GherkinParser, ParseResult
from gherlint.registry import CheckerRegistry
from gherlint.reporting import Message, TextReporter
from gherlint.walker import ASTWalker


class GherkinLinter(BaseChecker):
    """Main linter class which orchestrates the linting process."""

    MESSAGES = [
        Message("E001", "unparseable-file", "File could not be parsed: {error_msg}"),
        Message(
            "E002",
            "missing-language-tag",
            (
                "A feature file which uses an other language than English should declare "
                "this with a '# language: <lang>' tag at the beginning of the file."
            ),
        ),
        Message(
            "E003",
            "wrong-language-tag",
            "Language tag does not match the language used",
        ),
    ]

    def __init__(self, path: Path) -> None:
        super().__init__(reporter=TextReporter())
        self.path = path
        self.checker_registry = CheckerRegistry()
        self.checker_registry.discover()
        self.checkers: List[BaseChecker] = [
            checker(self.reporter) for checker in self.checker_registry
        ]
        self.walker = ASTWalker(self.checkers)

    def run(self) -> None:
        """Lint all feature files"""
        if self.path.is_file():
            self.lint_file(self.path)
        else:
            for filepath in self.path.rglob("*.feature"):
                self.lint_file(filepath)

    def lint_file(self, filepath: Path) -> None:
        result = GherkinParser().parse(filepath)
        if result.exception:
            self._handle_parser_error(result)
            return
        if result.added_language_tag:
            self.reporter.add_message(
                "missing-language-tag",
                result.document,
            )
        elif result.fixed_language_tag:
            self.reporter.add_message(
                "wrong-language-tag",
                result.document,
            )
        self.walker.walk(result.document)

    def _handle_parser_error(self, result: ParseResult) -> None:
        offending_lines = str(result.exception).splitlines()[1:]
        for offending_line in offending_lines:
            result.document.line, result.document.column, error_msg = parse.search(
                "({:d}:{:d}): {}, got", offending_line
            )
            self.reporter.add_message(
                "unparseable-file",
                result.document,
                error_msg=error_msg,
            )



import logging
from pathlib import Path

from gherlint.parser import GherkinParser
from gherlint.utils import iter_feature_files

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)


class LanguageFixer:
    def __init__(self, path: Path) -> None:
        self.path = path

    def run(self, modify: bool) -> None:
        if not modify:
            logger.warning("Dry run enabled! No files will be modified.")
        for file in iter_feature_files(self.path):
            parse_result = GherkinParser().parse(file)
            if parse_result.added_language_tag or parse_result.fixed_language_tag:
                logger.info(
                    "Patching %s, reason: %s language tag",
                    file,
                    "no" if parse_result.added_language_tag else "wrong",
                )
                if modify:
                    file.write_text(parse_result.content, "utf8")



from collections import Counter
from pathlib import Path
from typing import Tuple

from gherlint import utils
from gherlint.checkers.base_checker import BaseChecker
from gherlint.objectmodel import nodes
from gherlint.parser import GherkinParser
from gherlint.reporting import TextReporter
from gherlint.walker import ASTWalker


class Statistics(BaseChecker):
    counter: Counter = Counter()
    elements: Tuple[str, ...] = (
        "Features",
        "Backgrounds",
        "Scenario Outlines",
        "Examples",
        "Scenarios",
        "Steps",
        "Unparseable Files",
    )

    def visit_feature(self, _: nodes.Feature) -> None:
        self.counter.update(["Features"])

    def visit_background(self, _: nodes.Background) -> None:
        self.counter.update(["Backgrounds"])

    def visit_scenario(self, _: nodes.Scenario) -> None:
        self.counter.update(["Scenarios"])

    def visit_scenariooutline(self, _: nodes.ScenarioOutline) -> None:
        self.counter.update(["Scenario Outlines"])

    def visit_examples(self, node: nodes.Examples) -> None:
        self.counter.update(["Examples"] * node.number_of_entries)

    def visit_step(self, _: nodes.Step) -> None:
        self.counter.update(["Steps"])

    def print_summary(self) -> None:
        max_element_length: int = max(len(element) for element in self.elements)
        max_count_length: int = int(max(self.counter.values()) / 10)
        for element in self.elements:
            print("-" * (max_element_length + max_count_length + 7))
            print(
                f"| {element:<{max_element_length}} | {self.counter[element]:>{max_count_length}} |"
            )
        print("-" * (max_element_length + max_count_length + 7))


def compute_metrics(path: Path) -> None:
    parser = GherkinParser()
    statistics = Statistics(reporter=TextReporter())
    walker = ASTWalker(checkers=[statistics])
    for file in utils.iter_feature_files(path):
        result = parser.parse(file)
        if result.exception:
            statistics.counter.update(["Unparseable Files"])
        else:
            walker.walk(result.document)
    statistics.print_summary()



"""Parser for Gherkin feature files."""
from pathlib import Path
from typing import NamedTuple, Optional

import parse
from gherkin.dialect import DIALECTS
from gherkin.parser import CompositeParserException, Parser

from gherlint.objectmodel import nodes


class ParseResult(NamedTuple):
    document: nodes.Document
    content: str
    exception: Optional[CompositeParserException]
    added_language_tag: bool
    fixed_language_tag: bool


class GherkinParser:
    def __init__(self) -> None:
        self.parser = Parser()
        self.language = "en"
        self.content = ""
        self.offset = 0
        self.added_language_tag = False
        self.fixed_language_tag = False

    def parse(self, filepath: Path) -> ParseResult:
        self.content = filepath.read_text("utf8")
        self._detect_language()
        if self.language not in ("en", "unknown"):
            self._check_and_fix_language()
        try:
            data = self.parser.parse(self.content)
            data["filename"] = str(filepath)
            data["offset"] = self.offset
            document = nodes.Document.from_dict(data)
            exception = None
        except CompositeParserException as exc:
            document = nodes.Document(
                line=0, column=0, filename=str(filepath), feature=None, comments=[]
            )
            exception = exc
        return ParseResult(
            document=document,
            content=self.content,
            exception=exception,
            added_language_tag=self.added_language_tag,
            fixed_language_tag=self.fixed_language_tag,
        )

    def _detect_language(self) -> None:
        if "Feature:" in self.content:
            self.language = "en"
        for language, keywords in DIALECTS.items():
            if any(keyword in self.content for keyword in keywords["feature"]):
                self.language = language
                break
        else:
            self.language = "unknown"

    def _check_and_fix_language(self) -> None:
        match = parse.search("# language: {:l}\n", self.content)
        if not match:
            self.added_language_tag = True
            self.offset += 1
            self.content = f"# language: {self.language}\n" + self.content
        elif match[0] != self.language:
            self.fixed_language_tag = True
            self.content = self.content.replace(
                f"# language: {match[0]}", f"# language: {self.language}"
            )



import importlib
from pathlib import Path
from typing import Iterable, Iterator, List, Type

from gherlint.checkers.base_checker import BaseChecker

CHECKER_PACKAGE = "checkers"
CHECKER_PATH = Path(__file__).parent / CHECKER_PACKAGE
PREFIX = "gherlint"


class CheckerRegistry(Iterable):
    def __init__(self) -> None:
        self._checkers: List[Type[BaseChecker]] = []

    def discover(self) -> None:
        for modname in CHECKER_PATH.glob("*.py"):
            module = importlib.import_module(
                f"{PREFIX}.{CHECKER_PACKAGE}.{modname.stem}"
            )
            if hasattr(module, "register_checker"):
                module.register_checker(self)  # type: ignore

    def register(self, checker: Type[BaseChecker]) -> None:
        self._checkers.append(checker)

    def __iter__(self) -> Iterator[Type[BaseChecker]]:
        yield from self._checkers



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from gherlint.objectmodel import nodes


class GherlintException(Exception):
    """Base class for all custom exceptions issued by Gherlint"""


class InternalError(GherlintException):
    """Intended to be used in places where error handling is necessary e.g. for implementing
    guards to satisfy mypy etc. which should never be reached during normal execution.
    This exception class is never intended to reach the end user."""

    DEFAULT_MESSAGE = (
        "An internal error occured. This should never happen during normal operation, "
        "so if you weren't doing anything fancy, please file a bug report."
    )

    def __init__(self, node: Optional[nodes.Node], message: str = "") -> None:
        super().__init__()
        self.message = (
            self.DEFAULT_MESSAGE + "\n" + message if message else self.DEFAULT_MESSAGE
        )
        if node:
            self.message += (
                f" The error occured when processing the following node: {node}"
            )

    def __str__(self) -> str:
        return self.message


class DuplicateMessageError(GherlintException):
    """Raised if a message is already registered or either the id or name already exists."""


class UnknownMessageError(GherlintException):
    """Raised if an unknown message was requested from the MessageStore"""


class UnsupportedFiletype(GherlintException):
    """Raised if a file with unsupported extension should be processed."""



# pylint: disable=too-few-public-methods
from typing import List

from gherlint.reporting import Message, MessageStore, Reporter


class BaseChecker:
    """Base class for all checkers."""

    MESSAGES: List[Message] = []

    def __init__(self, reporter: Reporter):
        self.reporter = reporter
        for message in self.MESSAGES:
            MessageStore().register_message(message)



"""
Checker focussing on complexity issues.
These may be constructs that may be simplified, or are too long or complex which makes it harder to understand.
"""

from typing import Optional, Set, Union

from gherlint.checkers.base_checker import BaseChecker
from gherlint.objectmodel import nodes
from gherlint.registry import CheckerRegistry
from gherlint.reporting import Message, Reporter


class ComplexityChecker(BaseChecker):
    MESSAGES = [
        Message(
            "R201",
            "outline-could-be-a-scenario",
            "This outline contains no or only one example, consider using a normal scenario instead",
        ),
        Message(
            "R202",
            "consider-using-background",
            "Consider putting common 'Given' steps in a Background",
        ),
    ]

    def __init__(self, reporter: Reporter) -> None:
        super().__init__(reporter)
        self.common_given_steps: Optional[Set[str]] = None

    def visit_feature(self, _: nodes.Feature) -> None:
        self.common_given_steps = None

    def visit_scenario(
        self, node: Union[nodes.Scenario, nodes.ScenarioOutline]
    ) -> None:
        given_steps = set(
            step.text for step in node.steps if step.inferred_type == "given"
        )
        if self.common_given_steps is None:
            self.common_given_steps = given_steps
        else:
            self.common_given_steps &= given_steps

    def visit_scenariooutline(self, node: nodes.ScenarioOutline) -> None:
        self.visit_scenario(node)
        total_example_values = sum(
            example_set.number_of_entries for example_set in node.examples
        )
        if total_example_values < 2:
            self.reporter.add_message("outline-could-be-a-scenario", node)

    def leave_feature(self, node: nodes.Feature) -> None:
        if len(node.scenarios) > 1 and self.common_given_steps:
            for scenario in (child_node for child_node in node.scenarios):
                for step in scenario.steps:
                    if (
                        step.inferred_type == "given"
                        and step.text in self.common_given_steps
                    ):
                        self.reporter.add_message("consider-using-background", step)


def register_checker(registry: CheckerRegistry) -> None:
    registry.register(ComplexityChecker)



from typing import List

from gherlint.checkers.base_checker import BaseChecker
from gherlint.objectmodel.nodes import Node


class ASTWalker:
    """
    A class which walks through the inidvidual notes of an abstract syntax tree
    representing a Gherkin feature file and calls the callbacks.
    """

    def __init__(self, checkers: List[BaseChecker]) -> None:
        self.checkers = checkers

    def walk(self, node: Node) -> None:
        for checker in self.checkers:
            callback = getattr(
                checker, f"visit_{node.__class__.__name__.lower()}", None
            )
            if callback:
                callback(node)
        children = getattr(node, "children", ())
        for child_node in children:
            self.walk(child_node)
        for checker in self.checkers:
            callback = getattr(
                checker, f"leave_{node.__class__.__name__.lower()}", None
            )
            if callback:
                callback(node)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gherlint', 'gherlint.checkers', 'gherlint.objectmodel']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'gherkin-official>=22.0.0,<23.0.0',
 'parse>=1.19.0,<2.0.0']

entry_points = \
{'console_scripts': ['gherlint = gherlint.__main__:cli']}

setup_kwargs = {
    'name': 'gherlint',
    'version': '0.4.0',
    'description': 'Linter for Gherkin feature files',
    'long_description': '# gherlint\nLinter for Gherkin feature files, written in Python\n\n## Install\nSimply run ``pip install gherlint``.\n\n## Linting Feature Files\n``gherlint`` comes with a command line interface.\nTo recursively lint all feature files in a directory, run ``gherlint lint``.\n\n## Computing Metrics\n``gherlint`` can also create some metrics for you if you want to know how many features, scenarios and steps you have\nin your test suite. To do so, run ``gherlint stats``.\n\n## Disclaimer\n``gherlint`` is still in an early development phase. New checks will be added over time.\nIf you want to contribute, feel free to open issues suggesting useful checkers, or open a pull request if you want\nto!\n\n## Roadmap\n\nThe following work items are planned for the upcoming releases:\n\n* V0.0.x - V0.1.0:\n    * Parser and object model for feature files\n    * Basic checkers to demonstrate workflow\n    * Basic text based reporter\n* V0.1.x - V1.0: **<-- we are here**\n    * Add more checkers\n    * Add more output formats\n    * Extend object model as necessary for checkers\n    * Reach a stable interface for object model, checkers and reporters\n* V1.x - V2.0:\n    * Support configuration to enable/disable individual messages\n    * Implement plugin architecture to allow users to add custom checkers\n# CHANGELOG\n\n## V0.4.0\nNew checks:\n* ``unused-parameter``\n* ``empty-background``\n* ``outline-could-be-a-scenario``\n* ``consider-using-background``\n* ``examples-outside-scenario-outline``\n* ``duplicated-tag``\n\nOther changes:\n* ``unparseable-file`` is now issued for all nodes that offend the ``gherkin`` parser together with more error information provided by it\n* ``missing-parameter`` now clearly states what parameter is not defined in the examples\n* If ``gherlint`` patches the file contents with a forgotten ``# language`` tag the line numbers for messages\nare no longer shifted\n* Types of ``And`` and ``But`` steps are now inferred correctly\n* New command ``fix-language-tags`` to automatically add or fix missing or incorrect language tags\n* Rename some message codes to have a defined structure: each checker has its own hundreds digit.\n\n## V0.3.0\nNew checks:\n* ``missing-language-tag``\n* ``wrong-language-tag``\n* ``unparseable-file``\n\nOther changes:\n* ``gherlint`` can now automatically detect the language used and make sure that it can parse the files\neven without a ``# language`` token present.\n\n## V0.2.0\nNew checks:\n* ``missing-given-step``\n* ``missing-when-step``\n* ``missing-then-step``\n* ``empty-scenario``\n* ``empty-feature``\n* ``file-has-no-feature``\n* ``missing-parameter``\n\nOther changes:\n* Support for ``Background``\n* Determination of step type independent of language\n* Distinction between ``Scenario`` and ``Scenario Outline`` independent of language\n\n## V0.1.0\nFirst package in alpha status:\n* Parser and object model for feature files\n* Basic checkers to demonstrate workflow\n* Basic text based reporter\n',
    'author': 'Andreas Finkler',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DudeNr33/gherlint',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict

from gherlint.exceptions import DuplicateMessageError, UnknownMessageError
from gherlint.objectmodel.nodes import Document, Node


@dataclass(frozen=True)
class Message:
    id: str  # ID, like E001  pylint: disable=invalid-name
    name: str  # unique short name, like missing-feature-name
    text: str  # More detailed textual description of the problem and optionally a hint how to solve the problem.

    id_pattern = re.compile(r"^[EWCR]\d{3}$")
    name_pattern = re.compile(r"^[a-zA-Z]+(-[a-zA-Z]+)*$")

    def __post_init__(self):
        if not self.id_pattern.match(self.id):
            raise ValueError(f"Value for id must conform to {self.id_pattern.pattern}")
        if not self.name_pattern.match(self.name):
            raise ValueError(
                f"Value for name must conform to {self.name_pattern.pattern}"
            )


class MessageStore:
    id_to_message: Dict[str, Message] = {}
    name_to_message: Dict[str, Message] = {}

    @classmethod
    def register_message(cls, message: Message) -> None:
        if message.id in cls.id_to_message:
            raise DuplicateMessageError(
                f"Message with ID {message.id} already registered"
            )
        if message.name in cls.name_to_message:
            raise DuplicateMessageError(
                f"Message with name {message.name} already registered"
            )
        cls.id_to_message[message.id] = message
        cls.name_to_message[message.name] = message

    @classmethod
    def clear(cls):
        cls.id_to_message.clear()
        cls.name_to_message.clear()

    @classmethod
    def get_by_id(cls, message_id: str):
        try:
            return cls.id_to_message[message_id]
        except KeyError as exc:
            raise UnknownMessageError(f"Message ID {message_id} not found.") from exc

    @classmethod
    def get_by_name(cls, name: str):
        try:
            return cls.name_to_message[name]
        except KeyError as exc:
            raise UnknownMessageError(f"Message name '{name}' not found.") from exc


class Reporter(ABC):
    """Base class for reporters."""

    def add_message(self, id_or_name: str, node: Node, **format_args) -> None:
        """Add a message, identified by its id or name, that shall be emitted"""
        if Message.id_pattern.match(id_or_name):
            message = MessageStore.get_by_id(id_or_name)
        elif Message.name_pattern.match(id_or_name):
            message = MessageStore.get_by_name(id_or_name)
        else:
            raise ValueError(
                f"{id_or_name} matches neither the pattern for a message ID nor a message name."
            )
        self.emit(message, node, **format_args)

    @abstractmethod
    def emit(self, message: Message, node: Node, **format_args) -> None:
        """Emit the message as it is suitable for the desired report format"""


class TextReporter(Reporter):
    """Simple text based reporter that prints to stdout"""

    # example: missing_feature_name.feature:1:0: Feature has no name (missing-feature-name)
    MSG_TEMPLATE = "{file}:{line}:{column}: {text} ({name})"

    def __init__(self):
        self.current_file = None

    def emit(self, message: Message, node: Node, **format_args) -> None:
        root = node.get_root()
        if not isinstance(root, Document):
            raise RuntimeError(
                "The node passed to add_message does not have a root parent of type Document."
                "This should never happen if gherlint is used from the command line."
            )
        file = root.filename
        if file != self.current_file:
            self.current_file = file
            self.new_section_for_file()
        msg_text = message.text.format(**format_args) if format_args else message.text
        print(
            self.MSG_TEMPLATE.format(
                file=file,
                line=node.line,
                column=node.column,
                text=msg_text,
                name=message.name,
            )
        )

    def new_section_for_file(self):
        print(f"************* {self.current_file}")



"""
Checker focussing of completeness of feature files, e. g. that every scenario must have a name,
every parameter that is used in a scenario outline has to be defined in the examples, etc.
"""

from typing import Set, Union

from gherlint.checkers.base_checker import BaseChecker
from gherlint.exceptions import InternalError
from gherlint.objectmodel import nodes
from gherlint.registry import CheckerRegistry
from gherlint.reporting import Message, Reporter


class CompletenessChecker(BaseChecker):
    """Concerned about things like missing examples etc."""

    MESSAGES = [
        Message("W101", "missing-feature-name", "Feature has no name"),
        Message("W102", "missing-scenario-name", "Scenario has no name"),
        Message(
            "E101",
            "missing-parameter",
            "'{parameter}' is not defined in the Examples section",
        ),
        Message("W103", "file-has-no-feature", "No Feature given in file"),
        Message("W104", "empty-feature", "Feature has no scenarios"),
        Message("W105", "empty-scenario", "Scenario does not contain any steps"),
        Message("W106", "empty-background", "Background does not contain any steps"),
        Message(
            "C101", "missing-given-step", "Scenario does not contain any Given step"
        ),
        Message("C102", "missing-when-step", "Scenario does not contain any When step"),
        Message("C103", "missing-then-step", "Scenario does not contain any Then step"),
        Message("R101", "unused-parameter", "Parameter '{parameter}' is not used"),
    ]

    def __init__(self, reporter: Reporter) -> None:
        super().__init__(reporter)
        self.used_parameters: Set[str] = set()

    def visit_document(self, node: nodes.Document) -> None:
        if not node.feature:
            self.reporter.add_message("file-has-no-feature", node)

    def visit_feature(self, node: nodes.Feature) -> None:
        if not node.name.strip():
            self.reporter.add_message("missing-feature-name", node)
        if not node.children:
            self.reporter.add_message("empty-feature", node)

    def visit_background(self, node: nodes.Background) -> None:
        if not node.steps:
            self.reporter.add_message("empty-background", node)

    def visit_scenario(
        self, node: Union[nodes.Scenario, nodes.ScenarioOutline]
    ) -> None:
        if not node.name.strip():
            self.reporter.add_message("missing-scenario-name", node)
        if node.steps:
            self._check_missing_step_type(node)
        else:
            self.reporter.add_message("empty-scenario", node)

    def visit_scenariooutline(self, node: nodes.ScenarioOutline) -> None:
        # all checks relevant to normal scenarios apply for outlines as well
        self.visit_scenario(node)
        for param in node.parameters:
            self.used_parameters.add(param)
        self._check_missing_parameter(node)

    def leave_scenariooutline(self, _: nodes.ScenarioOutline) -> None:
        self.used_parameters.clear()

    def visit_examples(self, node: nodes.Examples) -> None:
        if isinstance(node.parent, nodes.ScenarioOutline):
            # Don't run this check if we are inside a normal scenario.
            # If examples are used we have a separate check for this (examples-outside-scenario-outline).
            self._check_unused_parameter(node)

    def visit_step(self, node: nodes.Step) -> None:
        if isinstance(node.parent, nodes.ScenarioOutline):
            for param in node.parameters:
                self.used_parameters.add(param)
            self._check_missing_parameter(node)

    def _check_missing_parameter(
        self, node: Union[nodes.ScenarioOutline, nodes.Step]
    ) -> None:
        if not node.parameters:
            return
        if isinstance(node, nodes.Step):
            if not isinstance(node.parent, nodes.ScenarioOutline):
                raise InternalError(node)
            outline = node.parent
        else:
            outline = node
        for param in node.parameters:
            if not any(param in examples.parameters for examples in outline.examples):
                self.reporter.add_message("missing-parameter", node, parameter=param)

    def _check_missing_step_type(
        self, node: Union[nodes.Scenario, nodes.ScenarioOutline]
    ):
        if not node.children:
            # We have an own message for this, don't need to clutter the output with
            # redundant information
            return
        required_step_types = ("given", "when", "then")
        for step_type in required_step_types:
            if not any(step.type == step_type for step in node.steps):
                self.reporter.add_message(f"missing-{step_type}-step", node)

    def _check_unused_parameter(self, node: nodes.Examples) -> None:
        for param in node.parameters:
            if param not in self.used_parameters:
                self.reporter.add_message("unused-parameter", node, parameter=param)


def register_checker(registry: CheckerRegistry) -> None:
    registry.register(CompletenessChecker)



from pathlib import Path
from typing import Iterator, List, Union

from gherkin.dialect import DIALECTS

from gherlint.exceptions import UnsupportedFiletype


def iter_feature_files(path: Union[str, Path]) -> Iterator[Path]:
    if isinstance(path, str):
        path = Path(path)
    if path.is_file():
        if path.suffix == ".feature":
            yield path
        else:
            raise UnsupportedFiletype(f"{path} is not a .feature file.")
    else:
        yield from path.rglob("*.feature")


def get_keyword_candidates(keyword: str) -> List[str]:
    """Get a list of the possible words of the keyword in all languages."""
    candidates = []
    for language in DIALECTS.values():
        candidates.extend(language[keyword])
    return candidates



"""
Nodes representing different elements of a Gherkin feature file.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import parse

from gherlint import utils
from gherlint.exceptions import InternalError


class Node(ABC):
    """
    Base class for all concrete node types.
    """

    def __init__(self, parent: Optional[Node], line: int, column: int):
        self.parent = parent
        self.line = line
        self.column = column
        root = self.get_root()
        if isinstance(root, Document) and root is not self:
            # shift the line number if the ``Document`` root node has an offset
            self.line -= root.offset

    def __repr__(self):
        return f"{self.__class__.__name__}(line={self.line}, column={self.column})"

    @property
    def parents(self) -> List[Node]:
        if self.parent:
            return [self.parent] + self.parent.parents
        return []

    def get_root(self) -> Node:
        """Get the root node, i.e. the topmost parent in the hierarchy."""
        current = self
        while current.parent is not None:
            current = current.parent
        return current

    @classmethod
    @abstractmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Node:
        """Create a node instance from the dictionary returned by the gherkin parser."""


class Document(Node):
    """Represents the file itself"""

    def __init__(
        self,
        line: int,
        column: int,
        filename: str,
        feature: Optional[Feature],
        comments: List[str],
        parent=None,
        offset: int = 0,
    ):
        super().__init__(parent, line, column)
        self.filename = filename
        self.feature = feature
        self.comments = comments
        # if we added a language tag we must shift all messages to refer to the correct line numbers
        self.offset = offset

    @property
    def children(self):
        return [self.feature]

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Node = None) -> Document:
        feature_data = data.get("feature")
        instance = cls(
            line=0,
            column=0,
            filename=data["filename"],
            feature=None,
            comments=data["comments"],
            offset=data.get("offset", 0),
        )
        if feature_data:
            instance.feature = Feature.from_dict(feature_data, parent=instance)
        return instance


class Feature(Node):
    """Represents a Feature in a file."""

    def __init__(
        self,
        line: int,
        column: int,
        parent: Optional[Node],
        tags: List[Tag],
        language: str,
        name: str,
        description: str,
        scenarios: List[Union[Scenario, ScenarioOutline]],
        background: Background = None,
    ):
        super().__init__(parent, line, column)
        self.tags = tags
        self.language = language
        self.name = name
        self.description = description
        self.background = background
        self.scenarios = scenarios

    @property
    def children(self) -> List[Union[Tag, Background, Scenario, ScenarioOutline]]:
        children: List[Union[Tag, Background, Scenario, ScenarioOutline]] = []
        children.extend(self.tags)
        if self.background:
            children.append(self.background)
        children.extend(self.scenarios)
        return children

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Feature:
        instance = cls(
            line=data["location"]["line"],
            column=data["location"]["column"],
            parent=parent,
            tags=[],
            language=data["language"],
            name=data["name"],
            description=data["description"],
            scenarios=[],
        )
        instance.tags = [Tag.from_dict(tag, parent=instance) for tag in data["tags"]]
        for child in data["children"]:
            for keyword, child_data in child.items():
                if _is_background(keyword, child_data):
                    instance.background = Background.from_dict(
                        child_data, parent=instance
                    )
                elif _is_scenario(keyword, child_data):
                    instance.scenarios.append(
                        Scenario.from_dict(child_data, parent=instance)
                    )
                elif _is_outline(keyword, child_data):
                    instance.scenarios.append(
                        ScenarioOutline.from_dict(child_data, parent=instance)
                    )
        return instance


def _is_scenario(keyword: str, data: Dict[str, Any]) -> bool:
    return keyword == "scenario" and data["keyword"] in utils.get_keyword_candidates(
        "scenario"
    )


def _is_outline(keyword: str, data: Dict[str, Any]) -> bool:
    return keyword == "scenario" and data["keyword"] in utils.get_keyword_candidates(
        "scenarioOutline"
    )


def _is_background(keyword: str, _: Dict[str, Any]) -> bool:
    return keyword == "background"


class Background(Node):
    """Represents a background of a feature."""

    def __init__(
        self,
        line: int,
        column: int,
        parent: Optional[Node],
        name: str,
        description: str,
        steps: List[Step],
    ):
        super().__init__(parent, line, column)
        self.name = name
        self.description = description
        self.steps = steps

    @property
    def children(self):
        return self.steps

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Background:
        instance = cls(
            line=data["location"]["line"],
            column=data["location"]["column"],
            parent=parent,
            name=data["name"],
            description=data["description"],
            steps=[],
        )
        instance.steps = [Step.from_dict(s, parent=instance) for s in data["steps"]]
        return instance


class Scenario(Node):
    """Represents a scenario of a feature."""

    def __init__(
        self,
        line: int,
        column: int,
        parent: Optional[Node],
        tags: List[Tag],
        name: str,
        description: str,
        examples: List[Examples],
        steps: List[Step],
    ):
        super().__init__(parent, line, column)
        self.tags = tags
        self.name = name
        self.description = description
        self.examples = examples
        self.steps = steps
        self.parameters = extract_parameters(name)

    @property
    def children(self):
        return self.tags + self.steps + self.examples

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Scenario:
        instance = cls(
            line=data["location"]["line"],
            column=data["location"]["column"],
            parent=parent,
            tags=[],
            name=data["name"],
            description=data["description"],
            examples=[],
            steps=[],
        )
        instance.tags = [Tag.from_dict(tag, instance) for tag in data["tags"]]
        instance.examples = [
            Examples.from_dict(d, parent=instance) for d in data["examples"]
        ]
        instance.steps = [Step.from_dict(s, parent=instance) for s in data["steps"]]
        return instance


class ScenarioOutline(Scenario):
    """Represents a scenario outline of a feature"""


class Step(Node):
    def __init__(
        self, parent: Optional[Node], line: int, column: int, keyword: str, text: str
    ):
        super().__init__(parent, line, column)
        self.type = self._get_english_keyword(keyword)
        self.text = text
        self.parameters = extract_parameters(text)

    @property
    def inferred_type(self) -> str:
        if not self.type in ("and", "but"):
            return self.type
        if not isinstance(self.parent, (Background, Scenario, ScenarioOutline)):
            raise InternalError(
                node=self, message="Unexpected type for parent of 'Step'"
            )
        current_step_index = self.parent.steps.index(self)
        while current_step_index > 0:
            current_step_index -= 1
            previous_step = self.parent.steps[current_step_index]
            if previous_step.type not in ("and", "but"):
                return previous_step.type
        return "unknown"

    @staticmethod
    def _get_english_keyword(keyword: str) -> str:
        """Get the corresponding english step keyword in lowercase from input in any (supported) language."""
        if keyword.strip() == "*":
            return "*"
        english_keywords = ("given", "when", "then", "and", "but")
        for english_keyword in english_keywords:
            if keyword.lower() in [
                kw.lower() for kw in utils.get_keyword_candidates(english_keyword)
            ]:
                return english_keyword
        raise ValueError(f"Unable to look up english step keyword for {keyword}")

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Step:
        return cls(
            line=data["location"]["line"],
            column=data["location"]["column"],
            parent=parent,
            keyword=data["keyword"],
            text=data["text"],
        )


class Examples(Node):
    def __init__(
        self,
        parent: Optional[Node],
        line: int,
        column: int,
        tags: List[Tag],
        name: str,
        description: str,
        parameters: List[str],
        values: Dict[str, List[str]],
    ):
        super().__init__(parent, line, column)
        self.tags = tags
        self.name = name
        self.description = description
        self.parameters = parameters
        self.values = values
        self.number_of_entries = len(values[parameters[0]])

    @property
    def children(self) -> List[Tag]:
        return self.tags

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Examples:
        parameters: List[str] = [cell["value"] for cell in data["tableHeader"]["cells"]]
        values: Dict[str, List[str]] = {param: [] for param in parameters}
        for row in data["tableBody"]:
            for param, entry in zip(parameters, row["cells"]):
                values[param].append(entry["value"])
        instance = cls(
            line=data["location"]["line"],
            column=data["location"]["column"],
            parent=parent,
            tags=[],
            name=data["name"],  # can this be filled?!
            description=data["description"],
            parameters=parameters,
            values=values,
        )
        instance.tags = [Tag.from_dict(tag, parent=instance) for tag in data["tags"]]
        return instance


class Tag(Node):
    def __init__(self, parent: Optional[Node], line: int, column: int, name: str):
        super().__init__(parent, line, column)
        self.name = name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Tag) and self.name == other.name

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parent: Optional[Node]) -> Tag:
        return cls(
            parent=parent,
            line=data["location"]["line"],
            column=data["location"]["column"],
            name=data["name"],
        )


def extract_parameters(text: str) -> Tuple[str]:
    """Extract parameters from a string (e. g. a step text).
    'Parameters' are placeholders defined in the Examples section of a
    Scenario Outline and are delimited with ``< >``."""
    pattern = "<{}>"
    return tuple(match.fixed[0] for match in parse.findall(pattern, text))  # type: ignore
